#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 uralbash <root@uralbash.ru>
#
# Distributed under terms of the MIT license.

"""
Routes for pyramid_pages
"""
from sqlalchemy import or_
from pyramid.events import BeforeRender
from pyramid.location import lineage
from pyramid.httpexceptions import HTTPNotFound

from . import CONFIG_MODELS, CONFIG_DBSESSION
from .security import HOME_PAGE, PREFIX_PAGE
from .resources import (
    BasePageResource,
    models_of_config,
    resource_of_node,
    resources_of_config
)


def add_globals(event):
    event['lineage'] = lineage


def page_factory(request):
    """ Page factory.

    Config models example:

    .. code-block:: python

        models = {
            '': [WebPage, CatalogResource],
            'catalogue': CatalogResource,
            'news': NewsResource,
        }
    """
    prefix = request.matchdict['prefix']  # /{prefix}/page1/page2/page3...
    settings = request.registry.settings
    dbsession = settings[CONFIG_DBSESSION]
    config = settings[CONFIG_MODELS]

    if prefix not in config:
        # prepend {prefix} to *traverse
        request.matchdict['traverse'] =\
            tuple([prefix] + list(request.matchdict['traverse']))
        prefix = None

    # Get all resources and models from config with the same prefix.
    resources = config.get(
        prefix, config.get(   # 1. get resources with prefix same as URL prefix
            '', config.get(   # 2. if not, then try to get empty prefix
                '/', None)))  # 3. else try to get prefix '/' otherwise None

    if not hasattr(resources, '__iter__'):
        resources = (resources, )

    tree = {}

    if not resources:
        return tree

    # Add top level nodes of resources in the tree
    for resource in resources:
        table = None
        if not hasattr(resource, '__table__')\
                and hasattr(resource, 'model'):
            table = resource.model
        else:
            table = resource

        if not hasattr(table, 'slug'):
            continue

        nodes = dbsession.query(table)
        if hasattr(table, 'parent_id'):
            nodes = nodes.filter(or_(
                table.parent_id == None,  # noqa
                table.parent.has(table.slug == '/')
            ))
        for node in nodes:
            if not node.slug:
                continue
            resource = resource_of_node(resources, node)
            tree[node.slug] = resource(node, prefix=prefix)
    return tree


def home_page_factory(request):
    settings = request.registry.settings
    dbsession = settings[CONFIG_DBSESSION]
    config = settings[CONFIG_MODELS]
    models = models_of_config(config)
    resources = resources_of_config(config)
    for table in models:
        if not hasattr(table, 'slug'):
            continue
        node = dbsession.query(table).filter(table.slug == '/').first()
        if node:
            return resource_of_node(resources, node)(node)
    raise HTTPNotFound


def register_views(*args):
    """ Registration view for each resource from config.
    """
    config = args[0]
    settings = config.get_settings()
    pages_config = settings[CONFIG_MODELS]
    resources = resources_of_config(pages_config)
    for resource in resources:
        if hasattr(resource, '__table__')\
                and not hasattr(resource, 'model'):
            continue
        resource.model.pyramid_pages_template = resource.template
        config.add_view(resource.view,
                        attr=resource.attr,
                        route_name=PREFIX_PAGE,
                        renderer=resource.template,
                        context=resource,
                        permission=PREFIX_PAGE)


def includeme(config):
    config.add_subscriber(add_globals, BeforeRender)

    # Home page factory
    config.add_route(HOME_PAGE, '/', factory=home_page_factory)
    config.add_view(BasePageResource.view,
                    attr=BasePageResource.attr,
                    route_name=HOME_PAGE,
                    renderer=BasePageResource.template,
                    context=BasePageResource,
                    permission=HOME_PAGE)

    # Default page factory
    config.add_route(PREFIX_PAGE, '/{prefix}*traverse', factory=page_factory)
    config.add_view(BasePageResource.view,
                    attr=BasePageResource.attr,
                    route_name=PREFIX_PAGE,
                    renderer=BasePageResource.template,
                    context=BasePageResource,
                    permission=PREFIX_PAGE)

    import pkg_resources
    pyramid_version = pkg_resources.get_distribution("pyramid").parsed_version
    if pyramid_version >= pkg_resources.SetuptoolsVersion('1.6a1'):
        # Allow you to change settings after including this function. This
        # fuature works only in version 1.6 or above.
        config.action('pyramid_pages_routes', register_views, args=(config, ))
    else:
        config.include(register_views)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 uralbash <root@uralbash.ru>
#
# Distributed under terms of the MIT license.

"""
Models for page.
"""
from sqlalchemy import (
    Column,
    String,
    Boolean,
    Integer,
    ForeignKey,
    UnicodeText
)
from sacrud.common import ClassProperty
from saexttype import SlugType, ChoiceType
from sqlalchemy.orm import foreign, relationship
from sqlalchemy_mptt import BaseNestedSets
from sqlalchemy.ext.declarative import declared_attr

REDIRECT_CHOICES = (
    ('200', 'OK (200)'),
    ('301', 'Moved Permanently (301)'),
    ('302', 'Moved Temporarily (302)'),
)


class PageMixin(object):

    name = Column(String, nullable=False)
    visible = Column(Boolean)
    in_menu = Column(Boolean)
    slug = Column(SlugType('name', False), unique=True, nullable=False)
    description = Column(UnicodeText)

    def __repr__(self):
        return self.name or ''

    def __json__(self, request):
        return self.id


class MpttPageMixin(BaseNestedSets, PageMixin):
    pass


class FlatPageMixin(PageMixin):
    pass


class RecursionPageMixin(PageMixin):
    """
    Model with single parent_id field
    """


class SeoMixin(object):

    seo_title = Column(String)
    seo_keywords = Column(String)
    seo_description = Column(String)
    seo_metatags = Column(UnicodeText)


class RedirectMixin(object):

    redirect_url = Column(String)
    redirect_type = Column(ChoiceType(choices=REDIRECT_CHOICES))

    @declared_attr
    def redirect_page(cls):
        return Column(
            Integer,
            ForeignKey(
                '{}.{}'.format(cls.__tablename__, cls.get_pk_name()),
                ondelete='CASCADE')
        )

    @declared_attr
    def redirect(cls):
        pk = getattr(cls, cls.get_pk_name())
        return relationship(
            cls, foreign_keys=[cls.redirect_page],
            remote_side='{}.{}'.format(cls.__name__, cls.get_pk_name()),
            primaryjoin=lambda: foreign(cls.redirect_page) == pk,
        )


class SacrudOptions(object):

    @ClassProperty
    def sacrud_detail_col(cls):
        options = [
            ('', [cls.name, cls.slug, cls.visible, cls.in_menu,
                  cls.description, getattr(cls, 'parent', None)])
        ]
        if all(hasattr(cls, name)
               for name in ('redirect_url', 'redirect', 'redirect_type')):
            options.append(
                ('Redirection', [cls.redirect_url, cls.redirect,
                                 cls.redirect_type])
            )
        if all(hasattr(cls, name)
               for name in ('seo_title', 'seo_keywords', 'seo_description',
                            'seo_metatags')):
            options.append(
                ('SEO', [cls.seo_title, cls.seo_keywords, cls.seo_description,
                         cls.seo_metatags])
            )
        return options


class BaseSacrudMpttPage(
        SacrudOptions,
        SeoMixin,
        RedirectMixin,
        MpttPageMixin
):
    """
    Base mptt page class for :mod:`pyramid_sacrud`.
    """


class BaseSacrudFlatPage(
        SacrudOptions,
        SeoMixin,
        RedirectMixin,
        FlatPageMixin
):
    """
    Base flat page class for :mod:`pyramid_sacrud`.
    """



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2015 uralbash <root@uralbash.ru>
#
# Distributed under terms of the MIT license.

"""
Base pages resources.
"""
import re

from pyramid.threadlocal import get_current_registry

from . import CONFIG_MODELS, CONFIG_DBSESSION
from .views import PageView


class BasePageResource(object):

    """ Base resource tree class for pages.

    :view: view of resource.
    :attr: The view machinery defaults to using the ``__call__`` method of the
        view callable (or the function itself, if the view callable is a
        function) to obtain a response. The attr value allows you to vary the
        method attribute used to obtain the response. For example, if your view
        was a class, and the class has a method named index and you wanted to
        use this method instead of the class's ``__call__`` method to return
        the response, you'd say attr="index" in the view configuration for the
        view. This is most useful when the view definition is a class.
    :template: template for view of resource.
    """

    view = PageView
    attr = 'page_with_redirect'
    template = 'pyramid_pages/index.jinja2'

    def __init__(self,
                 node, prefix=None, request=None, parent=None,
                 *args, **kwargs):
        """ Make resource of node for traversal URL dispath.

        :node: instance of page
        :pages_config: config for pyramid_pages from your app
        :resources: list of all resources in config
        :prefix: URL prefix for current node
        :parent: if exist it assignet to ``__parent__`` attribute
        """
        self.settings = get_current_registry().settings
        self.node = node
        self.dbsession = self.settings[CONFIG_DBSESSION]
        self.pages_config = self.settings[CONFIG_MODELS]
        self.resources = resources_of_config(self.pages_config)
        self.prefix = prefix or self.get_prefix()
        self.parent = parent

    def get_prefix(self):
        """ Each resource defined in config for pages as dict. This method
        returns key from config where located current resource.
        """
        for key, value in self.pages_config.items():
            if not hasattr(value, '__iter__'):
                value = (value, )
            for item in value:
                if type(self.node) == item\
                        or type(self.node) == getattr(item, 'model', None):
                    return key

    def resource_of_node(self, node, parent=None):
        return resource_of_node(self.resources, node)(
            node, parent=parent or self.parent)

    @property
    def __name__(self):
        """ Returns ``slug`` attribute of node or ``prefix``.
        """
        if self.node and not self.node.slug == '/':
            return self.node.slug
        elif self.node and self.node.slug == '/':
            return ''
        elif self.prefix:
            return self.prefix
        return None

    @property
    def name(self):
        return self.node.name

    @property
    def __parent__(self):
        if self.parent:
            return self.parent
        elif hasattr(self.node, 'parent'):
            parent = self.node.parent
            return resource_of_node(self.resources, parent)(parent)
        elif self.node:
            return self.__class__(None)
        return None

    def __getitem__(self, name):
        for child in self.children:
            if child.slug == name:
                return self.resource_of_node(child.node)

    @property
    def slug(self):
        return getattr(self.node, 'slug', None)

    @property
    def children(self):
        for child in self.node.children:
            yield self.resource_of_node(child)

    @property
    def children_qty(self):
        return len([child for child in self.children
                    if child.node.visible and child.node.in_menu])

    def __resource_url__(self, request, info):
        """ Some hook for prefix and doplication root slash.
        """
        separator = '/{}'.format(self.prefix) if self.prefix else ''
        # XXX: I feel a dissonance here
        info['virtual_path'] = re.sub('/+', '/', info['virtual_path'])
        return info['app_url'] + separator + info['virtual_path']

    def __repr__(self):
        return self.name


def resource_of_node(resources, node):
    """ Returns resource of node.
    """
    for resource in resources:
        model = getattr(resource, 'model', None)
        if type(node) == model:
            return resource
    return BasePageResource


def resources_of_config(config):
    """ Returns all resources and models from config.
    """
    return set(             # unique values
        sum([               # join lists to flat list
            list(value)     # if value is iter (ex: list of resources)
            if hasattr(value, '__iter__')
            else [value, ]  # if value is not iter (ex: model or resource)
            for value in config.values()
        ], [])
    )


def models_of_config(config):
    """ Return list of models from all resources in config.
    """
    resources = resources_of_config(config)
    models = []
    for resource in resources:
        if not hasattr(resource, '__table__') and hasattr(resource, 'model'):
            models.append(resource.model)
        else:
            models.append(resource)
    return models
